/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ...
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <VirtualRobot/Robot.h>

#include <RobotAPI/components/units/RobotUnit/RobotUnit.h>
#include <RobotAPI/components/units/RobotUnit/NJointControllers/NJointController.h>
#include <RobotAPI/components/units/RobotUnit/ControlTargets/ControlTarget1DoFActuator.h>
#include <RobotAPI/components/units/RobotUnit/SensorValues/SensorValue1DoFActuator.h>

#include <armarx/control/common/control_law/TaskspaceImpedanceController.h>
#include <armarx/control/interface/ConfigurableNJointControllerInterface.h>
#include <armarx/control/njoint_controller/task_space/ControllerInterface.h>
#include <armarx/control/njoint_controller/controller_descriptions.h>
#include <armarx/control/njoint_controller/task_space/aron/NJointBimanualControllerConfig.aron.generated.h>


namespace armarx::control::njoint_controller::task_space
{
    namespace law = armarx::control::common::control_law;

    struct TSBimanualImpedanceConfig:
    {
        control_law::TaskspaceImpedanceController::Config configLeft;
        control_law::TaskspaceImpedanceController::Config configRight;
    };

    /**
    * @defgroup Library-NJointTaskspaceImpedanceController NJointTaskspaceImpedanceController
    * @ingroup Library-RobotUnit-NJointControllers
    * A description of the library NJointTaskspaceImpedanceController.
    *
    * @class NJointTaskspaceImpedanceController
    * @ingroup Library-NJointTaskspaceImpedanceController
    * @brief Brief description of class NJointTaskspaceImpedanceController.
    *
    * Detailed description of class NJointTaskspaceImpedanceController.
    */
    class NJointTaskspaceBimanualImpedanceController :
            virtual public NJointController,
            virtual public NJointTaskspaceBimanualImpedanceControllerInterface
    {
    public:
        using ConfigPtrT = ConfigurableNJointControllerConfigPtr;
        using AronDTO = law::arondto::TaskspaceImpedanceControllerConfig;
        using BO = law::TaskspaceImpedanceController::Config;

        NJointTaskspaceImpedanceController(const RobotUnitPtr& robotUnit,
                                           const NJointControllerConfigPtr& config,
                                           const VirtualRobot::RobotPtr&);

        std::string getClassName(const Ice::Current&) const override;

        void rtRun(const IceUtil::Time& sensorValuesTimestamp,
                   const IceUtil::Time& timeSinceLastIteration) override;

        /// NJointController interface
        // std::string getKinematicChainName(const Ice::Current&) override;

        void updateConfig(const ::armarx::aron::data::dto::DictPtr& dto,
                          const Ice::Current& iceCurrent = Ice::emptyCurrent) override;

        void validateConfigData(BO& config);

    protected:
        virtual void additionalTask();
        void onPublish(const SensorAndControl&,
                       const DebugDrawerInterfacePrx&,
                       const DebugObserverInterfacePrx&) override;

    protected:
        /// devices
        struct SensorDevices
        {
            std::vector<const SensorValue1DoFActuatorTorque*> torqueSensors;
            std::vector<const SensorValue1DoFActuatorVelocity*> velocitySensors;
            std::vector<const SensorValue1DoFActuatorPosition*> positionSensors;
        };
        SensorDevices sensorDevicesLeft;
        SensorDevices sensorDevicesRight;

        std::vector<ControlTarget1DoFActuatorTorque*> targetsLeft;
        std::vector<ControlTarget1DoFActuatorTorque*> targetsRight;

    protected:
        /// variables
        std::string kinematicChainNameLeft;
        std::string kinematicChainNameRight;
        std::vector<std::string> jointNamesLeft;
        std::vector<std::string> jointNamesRight;
        law::TaskspaceImpedanceController controllerLeft;
        law::TaskspaceImpedanceController controllerRight;

        /// set buffers
        TripleBuffer<BO> bufferUserToAdditionalTask;
        TripleBuffer<BO> bufferUserToRt;
        TripleBuffer<law::RobotStatus> bufferRtToAdditionalTask;

        std::atomic_bool rtFirstRun = true;
        std::atomic_bool rtReady = false;
        std::atomic_bool reInitPreActivateLeft = false;
        std::atomic_bool reInitPreActivateRight = false;

        VirtualRobot::RobotPtr nonRtRobot;

        common::control_law::ControlTarget controlTargetsLeft;
        common::control_law::ControlTarget controlTargetsRight;

        RobotUnitPtr robotUnit;

    protected:
        /// NJointControllerBase interface
        void onInitNJointController() override;
        void rtPreActivateController() override;
        void rtPostDeactivateController() override;
    };
} /// namespace armarx::control::njoint_controller::task_space
