/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ImpedanceController.h"
#include <armarx/control/njoint_mp_controller/task_space/aron/aron_conversions.h>
#include <armarx/control/common/utils.h>


namespace armarx::control::njoint_mp_controller::task_space
{
    NJointControllerRegistration<NJointTSImpedanceMPController> registrationControllerNJointTSImpedanceMPController("NJointTSImpedanceMPController");


    NJointTSImpedanceMPController::NJointTSImpedanceMPController(
            const RobotUnitPtr& robotUnit,
            const NJointControllerConfigPtr& config,
            const VirtualRobot::RobotPtr& robot):
        NJointTaskspaceImpedanceController{robotUnit, config, robot},
        mp::MPPool{}
    {
        ConfigPtrT cfg = ConfigPtrT::dynamicCast(config);
        auto configData = ::armarx::fromAron<AronDTO, BO>(cfg->config);
        validateConfigData(configData);
        bufferUserToAdditionalTask.reinitAllBuffers(configData);
        bufferUserToRt.reinitAllBuffers(configData);
        ARMARX_IMPORTANT << VAROUT(configData.mpConfig.mpList.size());
        createMPs(configData.mpConfig);
        isMPReady.store(true);
        ARMARX_IMPORTANT << "NJointTSImpedanceMPController construction done";
    }


    std::string NJointTSImpedanceMPController::getClassName(const Ice::Current &) const
    {
        return "NJointTSImpedanceMPController";
    }


    void NJointTSImpedanceMPController::updateConfig(const ::armarx::aron::data::dto::DictPtr &dto, const Ice::Current &iceCurrent)
    {
        NJointTaskspaceImpedanceController::updateConfig(dto);
        if (isFinished("all"))
        {
            isMPReady.store(false);
            const auto& cfg = bufferUserToAdditionalTask.getUpToDateReadBuffer();
            reconfigureMPs(cfg.mpConfig);
            isMPReady.store(true);

            ARMARX_IMPORTANT << "reconfigure MP: reinitialize the mp input output, as well as the rt related buffer values";
            VirtualRobot::RobotNodeSetPtr rns = rtGetRobot()->getRobotNodeSet(kinematicChainName);
            Eigen::Matrix4f currentPose = rns->getTCP()->getPoseInRootFrame();
            for (auto& _mp: mps)
            {
                if (_mp.second.mp->getClassName() == "TSMP")
                {
                    ARMARX_IMPORTANT << "init input output data for " << _mp.second.mp->getMPName();
                    std::dynamic_pointer_cast<mp::TSMPInput>(_mp.second.input)->pose = currentPose;
                    std::dynamic_pointer_cast<mp::TSMPOutput>(_mp.second.output)->pose = currentPose;
                }
                else if (_mp.second.mp->getClassName() == "JSMP") {
                    ARMARX_IMPORTANT << "init input output data for " << _mp.second.mp->getMPName();
                    std::dynamic_pointer_cast<mp::JSMPInput>(_mp.second.input)->angleRadian = rns->getJointValuesEigen();
                }
            }
        }
    }


    void NJointTSImpedanceMPController::additionalTask()
    {
        //    NJointTaskspaceImpedanceController::additionalTask();
        robotUnit->updateVirtualRobot(nonRtRobot);

        auto& rt2Add = bufferRtToAdditionalTask.getUpToDateReadBuffer();
        auto& u2RtConfig = bufferUserToRt.getWriteBuffer();
        auto& u2AddConfig = bufferUserToAdditionalTask.getWriteBuffer();

        /// parse arguments and run mp
        if (isMPReady.load())
        {
            for (auto& _mp: mps)
            {
                if (_mp.second.mp->getClassName() == "TSMP")
                {
                    mp::TSMPInputPtr in = std::dynamic_pointer_cast<mp::TSMPInput>(_mp.second.input);
                    in->pose = controller.tcp->getPoseInRootFrame();
                    in->vel = controller.bufferNonRtToOnPublish.getWriteBuffer().currentTwist;
                    in->deltaT = rt2Add.deltaT;
                }
                else if (_mp.second.mp->getClassName() == "JSMP")
                {
                    mp::JSMPInputPtr in = std::dynamic_pointer_cast<mp::JSMPInput>(_mp.second.input);
                    for (size_t i = 0; i < rt2Add.jointPosition.size(); i++)
                    {
                        in->angleRadian(i) = rt2Add.jointPosition[i];
                        in->angularVel(i) = rt2Add.jointVelocity[i];
                    }
                    in->deltaT = rt2Add.deltaT;
                }
            }
            runMPs();
            /// set mp target to rt buffer
            for (auto& _mp: mps)
            {
                if (_mp.second.mp->getClassName() == "TSMP")
                {
                    mp::TSMPOutputPtr out = std::dynamic_pointer_cast<mp::TSMPOutput>(_mp.second.output);
                    u2RtConfig.desiredPose = out->pose;
                    u2RtConfig.desiredTwist = out->vel;
                    u2AddConfig.desiredPose = out->pose;
                    u2AddConfig.desiredTwist = out->vel;
                }
                else if (_mp.second.mp->getClassName() == "JSMP")
                {
                    mp::JSMPOutputPtr out = std::dynamic_pointer_cast<mp::JSMPOutput>(_mp.second.output);
                    u2RtConfig.desiredNullspaceJointAngles.value() = out->angleRadian;
                    u2AddConfig.desiredNullspaceJointAngles.value() = out->angleRadian;
                }
            }
            bufferUserToRt.commitWrite();
            bufferUserToAdditionalTask.commitWrite();
        }
        controller.updateControlStatus(u2AddConfig, rt2Add);
    }

    void NJointTSImpedanceMPController::rtPreActivateController()
    {
        ARMARX_IMPORTANT << "rt pre activate: reinitialize the mp input output, as well as the rt related buffer values";
        VirtualRobot::RobotNodeSetPtr rns = rtGetRobot()->getRobotNodeSet(kinematicChainName);
        Eigen::Matrix4f currentPose = rns->getTCP()->getPoseInRootFrame();
        NJointTaskspaceImpedanceController::rtPreActivateController();
        for (auto& _mp: mps)
        {
            if (_mp.second.mp->getClassName() == "TSMP")
            {
                ARMARX_IMPORTANT << "init input output data for " << _mp.second.mp->getMPName();
                std::dynamic_pointer_cast<mp::TSMPInput>(_mp.second.input)->pose = currentPose;
                std::dynamic_pointer_cast<mp::TSMPOutput>(_mp.second.output)->pose = currentPose;
            }
            else if (_mp.second.mp->getClassName() == "JSMP")
            {
                ARMARX_IMPORTANT << "init input output data for " << _mp.second.mp->getMPName();
                std::dynamic_pointer_cast<mp::JSMPInput>(_mp.second.input)->angleRadian = rns->getJointValuesEigen();
                std::dynamic_pointer_cast<mp::JSMPInput>(_mp.second.input)->angularVel =
                    Eigen::VectorXf::Zero(rns->getJointValues().size());
                std::dynamic_pointer_cast<mp::JSMPOutput>(_mp.second.output)->angleRadian =
                    rns->getJointValuesEigen();
                std::dynamic_pointer_cast<mp::JSMPOutput>(_mp.second.output)->angularVel =
                    Eigen::VectorXf::Zero(rns->getJointValues().size());
            }
        }
    }
} /// namespace armarx::control::njoint_mp_controller::task_space
